#region -- BSD Licence --
/*
Copyright (c) 2003, 5D Services
http://www.ihook.cc
OpenSource@ihook.cc
lloyd@5dservices.com.au
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. 

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution. 

Neither the name of "5D Services" nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
#endregion
using System;
using System.Drawing;
using System.Windows.Forms;

namespace ihook.PocketWidgets
{
	/// <summary>
	/// This is on off-screen buffer for double buffered control.
	/// it's tradeoff between performance and memory !...
	/// </summary>
	public class BackBuffer : IDisposable
	{
		int width, height;
		Bitmap image;
		Graphics graphics;

		/// <summary>
		/// create a new back buffer with a given size
		/// </summary>
		/// <param name="aWidth"></param>
		/// <param name="aHeight"></param>
		public BackBuffer(int aWidth, int aHeight)
		{
			width = aWidth;
			height = aHeight;
			if(width < 1)
				width = 1;
			if(height < 1)
				height = 1;
			init0();
		}
		~BackBuffer()
		{
			Dispose();
		}
		public void Dispose()
		{
			if(image == null)
				return;
			graphics.Dispose();
			image.Dispose();
			image = null;
			graphics = null;
			GC.SuppressFinalize(this);
		}
		void init()
		{
			if(image == null)
			{
				GC.ReRegisterForFinalize(this);
				init0();
			}
		}
		void init0()
		{
			image = new Bitmap(width, height);
			graphics = Graphics.FromImage(image);
		}
		void check()
		{
			if(image == null)
				throw new ObjectDisposedException("buffer destroyed, paint with graphics to recreate");
		}
		/// <summary>
		/// paint the buffer on screen 
		/// </summary>
		/// <param name="pea">the event arg to screen</param>
		public void Paint(PaintEventArgs pea)
		{
			check();
			pea.Graphics.DrawImage(image, 
				pea.ClipRectangle,
				pea.ClipRectangle,
				GraphicsUnit.Pixel);
		}

		/// <summary>
		/// get the graphics object with wich to draw to this buffer
		/// </summary>
		public Graphics Graphics 
		{ 
			get 
			{
				init();
				return graphics;
			}
		}

		public void ResetClip()
		{
			if(graphics != null)
				graphics.ResetClip();
		}

		/// <summary>
		/// get the bounding rect of this BackBuffer
		/// </summary>
		public Rectangle Bounds
		{
			get { return new Rectangle(0, 0, width, height); }
		}
		/// <summary>
		/// return the size of this BackBuffer
		/// </summary>
		public Size Size
		{
			get { return new Size(width, height); }
		}

		/// <summary>
		/// copy a part of this buffer to an other part of itself
		/// </summary>
		/// <param name="src"></param>
		/// <param name="dst"></param>
		public void CopyArea(Rectangle src, Rectangle dst)
		{
			check();
			graphics.ResetClip();
			graphics.DrawImage(image, dst, src, GraphicsUnit.Pixel);
		}
		/// <summary>
		/// copy a part of this buffer to some other surface
		/// </summary>
		/// <param name="src"></param>
		/// <param name="gdst"></param>
		/// <param name="dst"></param>
		public void CopyArea(Rectangle src, Graphics gdst, Rectangle dst)
		{
			check();
			gdst.DrawImage(image, dst, src, GraphicsUnit.Pixel);
		}

		/// <summary>
		/// convenience method to init the content of the buffer
		/// from an other
		/// </summary>
		/// <param name="other"></param>
		public void CopyBuffer(BackBuffer other)
		{
			check();
			Rectangle r = Bounds;
			other.Graphics.DrawImage(image, r, r, GraphicsUnit.Pixel);
		}
	}
}
