#region -- BSD Licence --
/*
Copyright (c) 2003, 5D Services
http://www.ihook.cc
OpenSource@ihook.cc
lloyd@5dservices.com.au
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. 

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution. 

Neither the name of "5D Services" nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
#endregion
using System;
using System.Drawing;
using System.Windows.Forms;

namespace ihook.PocketWidgets
{
	/// <summary>
	/// This is a control implementing the common functionality
	/// of double buffered control. it keeps a constant reference
	/// to its buffer, whic might lead to huge memory consumption.
	/// <see cref="DoubleBufferedPaintBackground"/>
	/// <see cref="DoubleBufferedPaint"/>
	/// </summary>
	/// <remarks>
	/// by default scrolling is disabled
	/// </remarks>
	public class DoubleBufferedControl : ScrollControl
	{
		public DoubleBufferedControl()
		{
			HScrollEnabled = false;
			VScrollEnabled = false;
		}
		protected override void Dispose(bool disposing)
		{
			base.Dispose (disposing);
			if(buffer != null)
			{
				buffer.Dispose();
				buffer = null;
			}
			if(cached != null)
			{
				cached.Dispose();
				cached = null;
			}
		}

		bool useBuffer = true;
		/// <summary>
		/// a flag stating wether or not the back buffer 
		/// should be used
		/// </summary>
		public bool DoubleBuffered
		{
			get { return useBuffer; }
			set
			{
				if(useBuffer == value)
					return;
				useBuffer = value;
				if(useBuffer)
					Refresh();
			}
		}

		/// <summary>
		/// the buffer where the drawing occur
		/// </summary>
		public virtual BackBuffer BackBuffer
		{
			get
			{
				if(buffer == null)
					buffer = new BackBuffer(Width, Height);
				return buffer;
			}
		}
		BackBuffer buffer;

		/// <summary>
		/// don't overwrite
		/// <see cref="DoubleBufferedPaintBackground"/>
		/// <see cref="DoubleBufferedPaint"/>
		/// </summary>
		/// <param name="e"></param>
		protected sealed override void OnPaintBackground(PaintEventArgs e)
		{
		}

		/// <summary>
		/// don't overwrite 
		/// <see cref="DoubleBufferedPaintBackground"/>
		/// <see cref="DoubleBufferedPaint"/>
		/// </summary>
		/// <param name="e"></param>
		protected override void OnPaint(PaintEventArgs e)
		{
			base.OnPaint (e);
			if(DoubleBuffered)
			{
				BackBuffer buf = BackBuffer;
				buf.ResetClip();
				Rectangle r = e.ClipRectangle;
				Graphics g = buf.Graphics;

				g.Clip = new Region(r);
				PaintEventArgs de = new PaintEventArgs(g, e.ClipRectangle);

				DoubleBufferedPaintBackground(de);
				DoubleBufferedPaint(de);

				buf.CopyArea(r, e.Graphics, r);
			}
			else
			{
				DoubleBufferedPaintBackground(e);
				DoubleBufferedPaint(e);
			}
		}
		/// <summary>
		/// the method to overwrite to customize your background,
		/// by default it paint a white background
		/// </summary>
		/// <param name="e"></param>
		protected virtual void DoubleBufferedPaintBackground(PaintEventArgs e)
		{
			base.OnPaintBackground(e);
		}
		/// <summary>
		/// the method to overwrite to customize your painting
		/// </summary>
		/// <param name="e"></param>
		protected virtual void DoubleBufferedPaint(PaintEventArgs e)
		{
		}

		protected override void OnParentChanged(EventArgs e)
		{
			base.OnParentChanged (e);
			if(cached != null)
			{
				cached.Dispose();
				cached = null;
			}
		}

		protected override void OnResize(EventArgs e)
		{
			base.OnResize(e);
			if(buffer != null)
			{
				int w = Width;
				int h = Height;
				if(w > 0 && h > 0)
				{
					BackBuffer buf2 = new BackBuffer(w, h);
					buf2.CopyBuffer(buffer);
					buffer.Dispose();
					buffer = buf2;
				}
				else
				{
					buffer.Dispose();
					buffer = null;
				}
			}
			if(cached != null)
			{
				cached.Dispose();
				cached = null;
			}
		}
		/// <summary>
		/// paint an area straight away, using double buffering, if needed
		/// </summary>
		/// <param name="r"></param>
		public void PaintArea(Rectangle r)
		{
			bool dbl = DoubleBuffered;
			Graphics g;
			if(dbl)
				g = BackBuffer.Graphics;
			else
				g = CachedGraphics;

			PaintArea(r, g);
			if(dbl)
				ApplyBuffer(r);
		}
		/// <summary>
		/// paint an area straight away, using the arg Graphics object
		/// </summary>
		/// <param name="r"></param>
		public void PaintArea(Rectangle r, Graphics g)
		{
			g.Clip = new Region(r);
			PaintEventArgs e = new PaintEventArgs(g, r);
			DoubleBufferedPaintBackground(e);
			DoubleBufferedPaint(e);
		}

		Graphics cached;
		public Graphics CachedGraphics
		{
			get
			{
				if(cached == null)
					cached = CreateGraphics();
				return cached;
			}
		}

		/// <summary>
		/// copy a portion of the buffer to the screen
		/// </summary>
		/// <param name="r">the rectangle to copy</param>
		public virtual void ApplyBuffer(Rectangle r)
		{
			if(buffer == null)
				return;
			CachedGraphics.Clip = new Region(r);
			BackBuffer.Graphics.ResetClip();
			BackBuffer.CopyArea(r, CachedGraphics, r);
		}
	}
}
