#region -- BSD Licence --
/*
Copyright (c) 2003, 5D Services
http://www.ihook.cc
OpenSource@ihook.cc
lloyd@5dservices.com.au
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. 

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution. 

Neither the name of "5D Services" nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
#endregion
using System;
using System.Windows.Forms;
using System.ComponentModel;

using ihook.PocketWidgets.Utils;

namespace ihook.PocketWidgets.Editors
{
	public enum NumType
	{
		Decimal,
		Single,
		Double,
		Byte,
		SByte,
		Short,
		UShort,
		Int,
		UInt,
		Long,
		ULong
	}

	/// <summary>
	/// Numeric data entry control
	/// </summary>
	public class NumEditor : TextBox, IEditor
	{
		public NumEditor(NumType inputType) : this()
		{
			InputType = inputType;
		}
		public NumEditor()
		{
			// set default input type
			InputType = NumType.Int;

			// NOTE: Existing context menu allows Paste command, with no known
			//	method to intercept. Only option is to reset to empty.
			//	(setting to Null doesn't work)
			this.ContextMenu = new ContextMenu();
			//Validated += new EventHandler(HValidate);
			TextChanged += new EventHandler(HValidate);
		}

		private NumType m_inpType;
		public NumType InputType
		{
			get{return m_inpType;}
			set 
			{ 
				m_inpType = value; 
				if(!IsValid(Text))
					Text = "";
			}
		}

		public override string Text
		{
			set
			{
				if(value == Text)
					return;
				if(value == "" || IsValid(value))
					base.Text = value;
			}
		}
		public virtual object Value
		{
			get 
			{ 
				string s = Text;
				if(s.Length == 0)
					return null;

				object numvalue = 0;
				try
				{
					switch(m_inpType)
					{
						case NumType.Decimal:
							numvalue = Convert.ToDecimal(s);
							break;
						case NumType.Single:
							numvalue = Convert.ToSingle(s);
							break;
						case NumType.Double:
							numvalue = Convert.ToDouble(s);
							break;
						case NumType.Byte:
							numvalue = Convert.ToByte(s);
							break;
						case NumType.SByte:
							numvalue = Convert.ToSByte(s);
							break;
						case NumType.UShort:
							numvalue = Convert.ToUInt16(s);
							break;
						case NumType.Short:
							numvalue = Convert.ToInt16(s);
							break;
						case NumType.UInt:
							numvalue = Convert.ToUInt32(s);
							break;
						case NumType.Int:
							numvalue = Convert.ToInt32(s);
							break;
						case NumType.ULong:
							numvalue = Convert.ToUInt64(s);
							break;
						case NumType.Long:
							numvalue = Convert.ToInt64(s);
							break;
					}
				}
				catch {}
				return numvalue;
			}
			set 
			{ 
				if(value == null || value == DBNull.Value)
				{
					Text = "";
				}
				else
				{
					string s = value.ToString();
					if(!IsValid(s))
						throw new ArgumentException();
					Text = value.ToString(); 
				}
			}
		}
		public bool IsNull
		{
			get { return Text == ""; }
			set
			{
				if(value)
					Text = "";
				else if(Text == "")
					Text = "0";
			}
		}

		private bool IsValid(string val)
		{
			// this method validates the ENTIRE string
			//	not each character
			bool ret = true;

			// handle first char == '-'
			if(val.Equals("-") 
				&& m_inpType != NumType.Byte 
				&& m_inpType != NumType.UShort
				&& m_inpType != NumType.UInt
				&& m_inpType != NumType.ULong)
				return ret;

			if(val.Equals("") || val.Equals(String.Empty))
				return true;

			// parse into dataType, errors indicate invalid value
			// NOTE: parsing also validates data type min/max
			try
			{
				switch(m_inpType)
				{
					case NumType.Single:
						float.Parse(val);
						break;
					case NumType.Double:
						double.Parse(val);
						break;
					case NumType.Decimal:
						decimal.Parse(val);
						break;
					case NumType.Byte:
						byte.Parse(val);
						break;
					case NumType.SByte:
						sbyte.Parse(val);
						break;
					case NumType.Short:
						short.Parse(val);
						break;
					case NumType.UShort:
						ushort.Parse(val);
						break;
					case NumType.Int:
						int.Parse(val);
						break;
					case NumType.UInt:
						uint.Parse(val);
						break;
					case NumType.Long:
						long.Parse(val);
						break;
					case NumType.ULong:
						ulong.Parse(val);
						break;
					default:
						ret = false;
						break;
				}
			}
			catch { ret = false; }
			return ret;
		}

		// verify it wrok and evetually change to do that on 
		// TextChanged...
		protected override void OnKeyPress(KeyPressEventArgs e)
		{
			// assemble new text with new KeyStroke
			//	and pass to validation routine.

			// NOTES;
			//	1) Delete key is NOT passed here
			//	2) control passed here after ProcessCmdKey() is run

			char c = e.KeyChar;
			if(!Char.IsControl(c))	// not sure about this?? nothing in docs about what is Control char??
			{
				// prevent spaces
				if(c.ToString() == " ")
				{
					e.Handled = true;
					return;
				}

				string newText = base.Text.Substring(0, base.SelectionStart)
					+ c.ToString() + base.Text.Substring(base.SelectionStart + base.SelectionLength);
			
				if(!IsValid(newText))
				{
					e.Handled = true;
					WinAPI.Beep();
				}
			}
			base.OnKeyPress(e);
		}

		public Control Control
		{
			get { return this; }
		}

		public event System.EventHandler ValueChanged;

		private void HValidate(object sender, EventArgs e)
		{
			if(ValueChanged != null)
				ValueChanged(this, EventArgs.Empty);
		}
	}
}