#region -- BSD Licence --
/*
Copyright (c) 2003, 5D Services
http://www.ihook.cc
OpenSource@ihook.cc
lloyd@5dservices.com.au
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. 

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution. 

Neither the name of "5D Services" nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
#endregion
using System;
using System.Drawing;
using System.Windows.Forms;

using ihook.PocketWidgets.Utils;

namespace ihook.PocketWidgets
{
	/// <summary>
	/// a sccrollable control, which automatically display
	/// ScrollBar on the side when it <c>ScrollableSize</c>
	/// get bigger than its Size. and there is also 
	/// a <c>ScrollableLocation</c> field wich is set by the scrollbar.
	/// this is not meant to host other control, just to be a big view
	/// with a small viewport.
	/// </summary>
	/// <remarks>
	/// this object is meant to be container for other control, and could
	/// fill itself automaticall from an other control, like a form 
	/// generated by the designer for example.
	/// </remarks>
	public class ScrollControl : SmartControl
	{
		public ScrollControl()
		{
			hscroll.Location = new Point(Width, 0);
			vscroll.Location = new Point(0, Height);
			scrollercorner.Location = new Point(Width, Height);
			Controls.Add(hscroll);
			Controls.Add(vscroll);
			Controls.Add(scrollercorner);
			scrollercorner.BackColor = AppDefaults.DisabledBackground;
			hscroll.ValueChanged += new EventHandler(HScrolled);
			vscroll.ValueChanged += new EventHandler(HScrolled);
		}
		public ScrollControl(Control c) : this()
		{
			Fill(c);
		}
		/// <summary>
		/// fill this control with the child control of the argument
		/// </summary>
		/// <param name="c"></param>
		public void Fill(Control body)
		{
			ClearControls();
			int maxX = 0, maxY = 0, tmp;
			foreach(Control c in body.Controls)
			{
				Controls.Add(c);
				tmp = c.Bounds.Right;
				if(maxX < tmp)
					maxX = tmp;
				tmp = c.Bounds.Bottom;
				if(maxY < tmp)
					maxY = tmp;
			}
			sLoc = new Point(0, 0);
			ScrollableSize = new Size(maxX, maxY);
			Size = body.ClientSize;
		}
		/// <summary>
		/// find a potential scrollable size from all the child control
		/// on this one, in case you add control manually
		/// </summary>
		public void CalculateScrollableSize()
		{
			int maxX = 0, maxY = 0, tmp;
			foreach(Control c in Controls)
			{
				if(c == hscroll 
					|| c == vscroll
					|| c == scrollercorner)
					continue;
				tmp = c.Bounds.Right;
				if(maxX < tmp)
					maxX = tmp;
				tmp = c.Bounds.Bottom;
				if(maxY < tmp)
					maxY = tmp;
			}
			ScrollableSize = new Size(maxX, maxY);
		}

		/// <summary>
		/// remove all child control, save the scroll bars
		/// </summary>
		public virtual void ClearControls()
		{
			Controls.Clear();
			Controls.Add(hscroll);
			Controls.Add(vscroll);
			Controls.Add(scrollercorner);
		}

		/// <summary>
		/// whether or not an horizontal scroll bar would be used
		/// </summary>
		public bool HScrollEnabled 
		{
			get { return enableHScroll; }
			set
			{
				if(value != enableHScroll)
				{
					enableHScroll = value;
					CheckScrollbars();
				}
			}
		}
		bool enableHScroll = true;

		/// <summary>
		/// whether or not a vertical scroll bar would be used
		/// </summary>
		public bool VScrollEnabled 
		{
			get { return enableVScroll; }
			set
			{
				if(value != enableVScroll)
				{
					enableVScroll = value;
					CheckScrollbars();
				}
			}
		}
		bool enableVScroll = true;

		// the corner in front of both scroll bar
		Control scrollercorner = new Control();

		/// <summary>
		/// the horizontal scroll bar used
		/// </summary>
		public HScrollBar HScrollBar { get { return hscroll; } }
		HScrollBar hscroll = new HScrollBar();

		/// <summary>
		/// the vertical scroll bar used
		/// </summary>
		public VScrollBar VScrollBar { get { return vscroll; } }
		VScrollBar vscroll = new VScrollBar();

		protected override void OnResize(EventArgs e)
		{
			base.OnResize(e);
			CheckScrollbars();
		}

		/// <summary>
		/// this method display the scroll bars if needed and set their 
		/// <c>Minimum</c>, <c>Maximum</c>, <c>LargeChange</c> value. 
		/// overwrite to set the <c>SmallChange</c>
		/// </summary>
		protected virtual void CheckScrollbars()
		{
			Size ws = Size;
			if(ws.Width<1 || ws.Height<1)
				return;
			Size wls = ScrollableSize;
			int eV = ws.Height;
			if(enableHScroll)
				eV -= hscroll.Height;
			int eH = ws.Width;
			if(enableVScroll)
				eH -= vscroll.Width;
			bool showH = eH < wls.Width && enableHScroll;
			bool showV = eV < wls.Height && enableVScroll;
			// fix a small error case
			if(wls.Width <= ws.Width && wls.Height <= ws.Height)
				showH = showV = false;

			if(showV)
			{
				vscroll.Location = new Point(ws.Width-vscroll.Width, 0);
				vscroll.Height = ws.Height;
				if(showH)
					vscroll.Height -= hscroll.Height;
				vscroll.Visible = true;

				vscroll.Minimum = 0;
				vscroll.Maximum = wls.Height;
				vscroll.LargeChange = eV;
				vscroll.SmallChange = 20;
				if(vscroll.Value > wls.Height-eV) 
				{
					int newVal = wls.Height-eV;
					if(newVal < 0)
						newVal = 0;
					vscroll.Value = newVal;
				}
			}
			else
			{
				vscroll.Visible = false;
				vscroll.Value = 0;
			}

			if(showH)
			{
				hscroll.Location = new Point(0, ws.Height-hscroll.Height);
				hscroll.Width = ws.Width;
				if(showV)
					hscroll.Width -= vscroll.Width;
				hscroll.Visible = true;

				hscroll.Minimum = 0;
				hscroll.Maximum = wls.Width;
				hscroll.LargeChange = eH;
				hscroll.SmallChange = 20;
				if(hscroll.Value > wls.Width-eH)
				{
					int newVal = wls.Width-eH;
					if(newVal < 0)
						newVal = 0;
					hscroll.Value = newVal;
				}
			}
			else
			{
				hscroll.Visible = false;
				hscroll.Value = 0;
			}
			if(showV && showH)
			{
				scrollercorner.Size = new Size(vscroll.Width, hscroll.Height);
				scrollercorner.Location = new Point(vscroll.Location.X, hscroll.Location.Y);
				scrollercorner.Visible = true;
			}
			else
				scrollercorner.Visible = false;
		
		}

		public virtual void ScrollToVisible(Rectangle r)
		{
			Size sSize = ScrollableSize;
			r.Intersect(new Rectangle(0,0,sSize.Width,sSize.Height));

			Size view = Size;
			if(hscroll.Visible)
				view.Height -= hscroll.Height;
			if(vscroll.Visible)
				view.Width -= hscroll.Width;

			Point p = ScrollableLocation;
			Point p2 = p;
			ScrollPane.ScrollToRectVisible(view, r, ref p);

			if(p2.X != p.X || p2.Y != p.Y)
				ScrollableLocation = p;
		}
		/// <summary>
		/// caculate the new origin fo viewing when you want to do a minimal 
		/// scroll to show a given rect
		/// </summary>
		/// <param name="viewport">the viewing rectangle, something like 
		/// <c>0, 0, Width, Height</c></param>
		/// <param name="cell">the rectangle you want to view</param>
		/// <param name="origin">the current ScrollableLocation of your ScrollPane, 
		/// and fill it with the desired location</param>
		public static void ScrollToRectVisible(Size viewport, Rectangle cell, ref Point origin)
		{
			Rectangle view = new Rectangle(- origin.X, - origin.Y, viewport.Width, viewport.Height);

			// X - scrolling
			if(cell.Width > view.Width)
			{
				if(cell.X > view.X)
					view.X = cell.X;
				else if(cell.Right < view.Right)
					view.X = cell.Right - view.Width;
			}
			else
			{
				if(cell.X < view.X)
					view.X = cell.X;
				else if(cell.Right > view.Right)
					view.X = cell.Right - view.Width;
			}

			// Y - scrolling
			if(cell.Height > view.Height)
			{
				if(cell.Y > view.Y)
					view.Y = cell.Y;
				else if(cell.Bottom < view.Bottom)
					view.Y = cell.Bottom - view.Height;
			}
			else
			{
				if(cell.Y < view.Y)
					view.Y = cell.Y;
				else if(cell.Bottom > view.Bottom)
					view.Y = cell.Bottom - view.Height;
			}

			origin.X = - view.X;
			origin.Y = - view.Y;
		}

		private void HScrolled(object sender, EventArgs e)
		{
			Point p = new Point(- hscroll.Value, - vscroll.Value);
			ScrollableLocation = p;
		}

		/// <summary>
		/// return the viewing rectangle less the space used by the
		/// scrollbars (if any)
		/// </summary>
		public Rectangle InsideRect
		{
			get 
			{
				Rectangle r = new Rectangle(0, 0, Width, Height);
				if(hscroll.Visible)
					r.Height -= hscroll.Height;
				if(vscroll.Visible)
					r.Width -= vscroll.Width;
				return r;
			}
		}

		Point sLoc;
		/// <summary>
		/// the "logical location" of this view compare to the 
		/// real widget, kind of viewport hole.
		/// </summary>
		/// <remarks>this accessor will call <c>MovedBy()</c> 
		/// if it has been moved</remarks>
		public virtual Point ScrollableLocation
		{
			get { return sLoc; }
			set 
			{
				Point p0 = sLoc;
				Point p1 = value;
				int dx = p1.X-p0.X;
				int dy = p1.Y-p0.Y;

				if(dx!=0 || dy!=0)
				{
					ApplicationEx.DoEvents(WinAPI.WM_PAINT);
					//ApplicationEx.ClearCurrent();
					sLoc = value; 
					hscroll.Value = - sLoc.X;
					vscroll.Value = - sLoc.Y;
					MovedBy(dx, dy);
					if(Scrolled != null)
						Scrolled(this, EventArgs.Empty);
				}
			}
		}

		/// <summary>
		/// triggered whenever the ScrollableLocation has changed
		/// </summary>
		public event EventHandler Scrolled;


		/// <summary>
		/// this method is called by ScollableLocation 
		/// when it has changed, it's responsible for moving data
		/// and/or repainting.
		/// <see cref="ScrollArea"/>
		/// </summary>
		/// <param name="dx"></param>
		/// <param name="dy"></param>
		protected virtual void MovedBy(int dx, int dy)
		{
			Rectangle r = new Rectangle(0, 0, Width, Height);
			foreach(Control c in Controls)
			{
				if(c == scrollercorner
					|| c == hscroll
					|| c == vscroll)
					continue;
				Point p = c.Location;
				c.Location = new Point(p.X+dx, p.Y+dy);
			}
		}

		Size sSize;
		public virtual Size ScrollableSize
		{
			get { return sSize; }
			set 
			{
				sSize = value;
				CheckScrollbars();
			}
		}

		protected override void OnMouseDown(MouseEventArgs e)
		{
			base.OnMouseDown (e);
			// 'if' because this event could trigger a remove ...
			if(TopLevelControl is Form 
				&& TopLevelControl.Visible 
				&& Enabled
				&& ! OwnFocus)
				Focus();
		}

		/// <summary>
		/// this key handler is called whenever key event are send 
		/// to the control or its scroll bar
		/// </summary>
		protected override void OnKeyDown(KeyEventArgs e)
		{
			base.OnKeyDown(e);
			ApplicationEx.ClearCurrent();
			switch(e.KeyCode)
			{
				case Keys.Up:
					if(vscroll.Visible)
						vscroll.Value -= vscroll.SmallChange;
					break;
				case Keys.Down:
					if(vscroll.Visible)
						vscroll.Value += vscroll.SmallChange;
					break;
				case Keys.Left:
					if(hscroll.Visible)
						hscroll.Value -= hscroll.SmallChange;
					break;
				case Keys.Right:
					if(hscroll.Visible)
						hscroll.Value += hscroll.SmallChange;
					break;
			}
		}
	}
}
