#region -- BSD Licence --
/*
Copyright (c) 2003, 5D Services
http://www.ihook.cc
OpenSource@ihook.cc
lloyd@5dservices.com.au
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. 

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution. 

Neither the name of "5D Services" nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
#endregion
using System;
using System.Drawing;
using System.Windows.Forms;

namespace ihook.PocketWidgets
{
	/// <summary>
	/// an implementation of a scollable control which is not meant to
	/// display child control bug just a small part of big view (itself)
	/// <see cref="ScrollableSize"/>
	/// <see cref="ScrollableLocation"/>
	/// <see cref="MovedBy"/>
	/// </summary>
	public class ScrollPane : DoubleBufferedControl
	{
		public ScrollPane()
		{
			HScrollEnabled = true;
			VScrollEnabled = true;
		}

		/// <summary>
		/// calculcate all rectangles involved in a scrolling, to do 
		/// a minimal redraw
		/// </summary>
		/// <param name="frame">the area in which scrolling calculation
		/// are meant to occur</param>
		/// <param name="dp">the deplacement of the frame</param>
		/// <param name="srcCopy">which portion should be copyed</param>
		/// <param name="dstCopy">where it should be copied</param>
		/// <param name="dirtyArea">dirty areas left to repaint</param>
		public static void ScrollInfo(Rectangle frame, Point dp, 
			out Rectangle srcCopy, 
			out Rectangle dstCopy,
			out Rectangle[] dirtyArea)
		{
			if(dp.X == 0 && dp.Y == 0)
			{
				srcCopy = frame;
				dstCopy = frame;
				dirtyArea = new Rectangle[0];
				return;
			}

			Rectangle r = frame;
			r.X += dp.X;
			r.Y += dp.Y;
			r.Intersect(frame);
			dstCopy = r;
			r.X -= dp.X;
			r.Y -= dp.Y;
			srcCopy = r;

			if(srcCopy.Width == 0 || srcCopy.Height == 0)
			{
				dirtyArea = new Rectangle[] { frame };
			}
			else if(dp.Y == 0)
			{
				dirtyArea = new Rectangle[1];
				dirtyArea[0].Height = frame.Height;
				dirtyArea[0].Y = frame.Y;
				if(dp.X > 0) 
				{
					dirtyArea[0].X = frame.X;
					dirtyArea[0].Width = dp.X;
				}
				else
				{
					dirtyArea[0].X = frame.Right + dp.X;
					dirtyArea[0].Width = - dp.X;
				}
			}
			else if(dp.X == 0)
			{
				dirtyArea = new Rectangle[1];
				dirtyArea = new Rectangle[1];
				dirtyArea[0].Width = frame.Width;
				dirtyArea[0].X = frame.X;
				if(dp.Y > 0) 
				{
					dirtyArea[0].Y = frame.Y;
					dirtyArea[0].Height = dp.Y;
				}
				else
				{
					dirtyArea[0].Y = frame.Bottom + dp.Y;
					dirtyArea[0].Height = - dp.Y;
				}
			}
			else
			{
				dirtyArea = new Rectangle[2];

				dirtyArea[0].Width = frame.Width;
				dirtyArea[0].X = frame.X;
				if(dp.Y > 0) 
				{
					dirtyArea[0].Y = frame.Y;
					dirtyArea[0].Height = dp.Y;

					dirtyArea[1].Y = frame.Y + dp.Y;
					dirtyArea[1].Height = frame.Height - dp.Y;
				}
				else
				{
					dirtyArea[0].Y = frame.Bottom + dp.Y;
					dirtyArea[0].Height = - dp.Y;

					dirtyArea[1].Y = frame.Y;
					dirtyArea[1].Height = frame.Height + dp.Y;
				}

				if(dp.X > 0) 
				{
					dirtyArea[1].X = frame.X;
					dirtyArea[1].Width = dp.X;
				}
				else
				{
					dirtyArea[1].X = frame.Right + dp.X;
					dirtyArea[1].Width = - dp.X;
				}
			}
		}

		/// <summary>
		/// this method is called when the user has scrolled.
		/// it provide a default repaint implementation
		/// <see cref="ScrollArea"/>
		/// </summary>
		/// <param name="dx"></param>
		/// <param name="dy"></param>
		protected override void MovedBy(int dx, int dy)
		{
			base.MovedBy(dx, dy); 
			Rectangle r = new Rectangle(0, 0, Width, Height);
			ScrollArea(r, new Point(dx, dy));
		}

		/// <summary>
		/// flag stating wether or not scrolling is just a simple
		/// redraw of if we should BitBlt/Copy common area and
		/// redraw as less as possible
		/// </summary>
		bool OptimizedScrolling = true;

		/// <summary>
		/// this method is to be called once the ScrollLocation
		/// has been modified, it optimised the drawing speed by
		/// doing BitBlt of the screen of copy of back buffer
		/// </summary>
		/// <remarks>this method don't apply the buffer</remarks>
		/// <param name="r">the zone where the scrolling occur</param>
		/// <param name="dp">the deplacement</param>
		public virtual void ScrollArea(Rectangle r, Point dp)
		{
			if(!OptimizedScrolling)
			{
				PaintArea(r);
				return;
			}

			Rectangle srcCopy, dstCopy;
			Rectangle[] dirty;
			ScrollInfo(r, dp,
				out srcCopy, out dstCopy,
				out dirty);

			Graphics g;
			bool dbl = DoubleBuffered;
			if(dbl)
				g = BackBuffer.Graphics;
			else
				g = CachedGraphics;

			if(dbl)
				BackBuffer.CopyArea(srcCopy, dstCopy);
			else
				CopyArea(srcCopy, dstCopy);

			for(int i=0; i<dirty.Length; i++)
				PaintArea(dirty[i], g);
			if(dbl)
				ApplyBuffer(r);
		}
	}
}
