#region -- BSD Licence --
/*
Copyright (c) 2003, 5D Services
http://www.ihook.cc
OpenSource@ihook.cc
lloyd@5dservices.com.au
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. 

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution. 

Neither the name of "5D Services" nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
#endregion
using System;
using System.Drawing;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;

using ihook.PocketWidgets.Utils;

namespace ihook.PocketWidgets
{
	/// <summary>
	/// A control with additional feature like context menu,
	/// etc... based on WinAPI functions
	/// </summary>
	public class SmartControl : Control
	{
		public SmartControl()
		{
		}

		public static MenuItem CreateMenuItem(string title, EventHandler h)
		{
			MenuItem mi = new MenuItem();
			mi.Text = title;
			if(h != null)
				mi.Click += h;
			return mi;
		}

#if !NETCFDESIGNTIME
		IntPtr handle = IntPtr.Zero;
		public IntPtr Handle
		{
			get 
			{
				if(handle.Equals(IntPtr.Zero))
					handle = WinAPI.GetHandle(this);
				return handle;
			}
		}
#endif

		/// <summary>
		/// return wether or not this control of one of its child is focused...
		/// </summary>
		public bool OwnFocus
		{
			get { return TestFocusHierarchy(this); }
		}
		static bool TestFocusHierarchy(Control c)
		{
			if(c.Focused)
				return true;
			foreach(Control c2 in c.Controls)
				if(TestFocusHierarchy(c2))
					return true;
			return false;
		}

		/// <summary>
		/// provide a way to invoke method after the current event,
		/// as the Invoke method call it straight away if called in the
		/// main thread.
		/// </summary>
		/// <param name="eh"></param>
		protected void InvokeLater(EventHandler eh)
		{
			ThreadPool.QueueUserWorkItem(new WaitCallback(InvokeNow), eh);
		}
		void InvokeNow(object state)
		{
			Invoke((EventHandler) state);
		}

		public Point TranslatePoint(Point p, Control to)
		{
			return TranslatePoint(p, this, to);
		}
		public static Point TranslatePoint(Point p, Control from, Control to)
		{
			p = from.PointToScreen(p);
			return to.PointToClient(p);
		}
		/// <summary>
		/// test if the current mouse event is a context menu event.
		/// display the small red circle, to be called in onMouseDown
		/// </summary>
		/// <param name="e">the event to test</param>
		/// <returns>wether or not a context menu event occured</returns>
		public bool IsContextMenu(MouseEventArgs e)
		{
			// just to be sure
			// should never happened on a PPC
			if(e.Button == MouseButtons.Right)
				return true;

			WinAPI.SHRGINFO shrginfo = new WinAPI.SHRGINFO();
			shrginfo.cbSize = (uint) Marshal.SizeOf(shrginfo);
			shrginfo.hwndClient = Handle;
			shrginfo.x = e.X;
			shrginfo.y = e.Y;
			shrginfo.dwFlags = WinAPI.SHRG_RETURNCMD;
		
			if (WinAPI.SHRecognizeGesture(shrginfo) == WinAPI.GN_CONTEXTMENU)
				return true;
			return false;
		}
		/// <summary>
		/// test if the current mouse event is a context menu event.
		/// display the small red circle, to be called in onMouseDown
		/// </summary>
		/// <param name="e">the event to test</param>
		/// <returns>wether or not a context menu event occured</returns>
		public static bool IsContextMenu(Control c, MouseEventArgs e)
		{
			// just to be sure
			// should never happened on a PPC
			if(e.Button == MouseButtons.Right) 
				return true;

			WinAPI.SHRGINFO shrginfo = new WinAPI.SHRGINFO();
			shrginfo.cbSize = (uint) Marshal.SizeOf(shrginfo);
			shrginfo.hwndClient = WinAPI.GetHandle(c);
			shrginfo.x = e.X;
			shrginfo.y = e.Y;
			shrginfo.dwFlags = WinAPI.SHRG_RETURNCMD;
		
			if (WinAPI.SHRecognizeGesture(shrginfo) == WinAPI.GN_CONTEXTMENU)
				return true;
			return false;
		}

		public void XORPaint(Rectangle r, Color c)
		{
			FillColor(Handle, r, WinAPI.RasterOp.DSTINVERT, c);
		}
		public void FillColor(Rectangle r, WinAPI.RasterOp op, Color c)
		{
			FillColor(Handle, r, op, c);
		}
		public static void FillColor(IntPtr hWnd, Rectangle r, WinAPI.RasterOp op, Color c)
		{
			IntPtr newb = WinAPI.CreateSolidBrush(c);
			if(newb.Equals(IntPtr.Zero))
				throw new System.ComponentModel.Win32Exception();
			IntPtr dc = WinAPI.GetDC(hWnd);
			try 
			{
				IntPtr oldb = WinAPI.SelectObject(dc, newb);
				if(oldb.Equals(IntPtr.Zero))
				{
					WinAPI.DeleteObject(newb);
					throw new System.ComponentModel.Win32Exception();
				}
				try
				{
					if(! WinAPI.PatBlt(dc, 
						r.Left, r.Top, r.Width, r.Height,
						op))
						throw new System.ComponentModel.Win32Exception();
				}
				finally 
				{
					WinAPI.SelectObject(dc, oldb);
					WinAPI.DeleteObject(newb);
				}
			}
			finally { WinAPI.ReleaseDC(dc); }
		}
		/// <summary>
		/// copy an area of this control on some other place of itself.
		/// usefull for fast scrolling
		/// </summary>
		/// <param name="src"></param>
		/// <param name="dst"></param>
		public void CopyArea(Rectangle src, Rectangle dst)
		{
			IntPtr dc = WinAPI.GetDC(this);
			try
			{
				WinAPI.BitBlt(dc,
					dst.X, dst.Y, dst.Width, dst.Height,
					dc,
					src.X, src.Y, 
					WinAPI.RasterOp.SRCCOPY);
			}
			finally { WinAPI.ReleaseDC(dc); }
		}
	}
}
