#region -- BSD Licence --
/*
Copyright (c) 2003, 5D Services
http://www.ihook.cc
OpenSource@ihook.cc
lloyd@5dservices.com.au
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. 

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution. 

Neither the name of "5D Services" nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
#endregion
using System;
using System.Collections;
using System.Data;
using System.Drawing;
using System.Windows.Forms;

using ihook.PocketWidgets;
using ihook.PocketWidgets.Utils;
using ihook.PocketWidgets.Table.Renderers;

namespace ihook.PocketWidgets.Table
{
	public delegate void TableClickEventHandler(object source, TableClickEvent tce);
	public class TableClickEvent : EventArgs
	{
		public TableClickEvent(MouseEventArgs e, TableControl tc, Rectangle rect, int r, int c)
		{
			ITableModel tm = tc.Model;
			int nCol = tm==null ? 0 : tm.NumColumns;
			int nRow = tm==null ? 0 : tm.NumRows;
			Row = r<-1||r>=nRow ? -2 : r;
			Col = c<-1||c>=nCol ? -2 : c;
			if(!tc.ShowRowHeader && Row<0)
				Row = -2;
			if(!tc.ShowColHeader && Col<0)
				Col = -2;
			CellRectangle = rect;
			MouseDown = new Point(e.X, e.Y);
			Target = tc;
			MouseEvent = e;
		}
		public readonly TableControl Target;
		public readonly Point MouseDown;
		public readonly Rectangle CellRectangle;
		public int Row;
		public int Col;
		public MouseEventArgs MouseEvent;
		
		bool consumed = false;
		public void Consume() { consumed = true; }
		/// <summary>
		/// a consumed mousedown event won't generate
		/// the TableClickUp event and won't generate an editing
		/// </summary>
		public bool Consumed { get { return consumed; } }

		public bool IsContextMenu()
		{
			return Target.IsContextMenu(MouseEvent);
		}
	}

	/// <summary>
	/// a Table control which is completely owner drawn, and editable.
	/// to change editor look at the <see cref="EditorManager"/> property.
	/// to change the width or height of the cells programatically look at 
	/// <see cref="Body"/> property. to set the data which would be displayed
	/// look at the <see cref="Model"/> property.
	/// To change the way things are displayed look at the various 
	/// <c>xxxRenderer</c> or <c>RendererXXX</c> methods and properties.
	/// </summary>
	/// <remarks>this table only support single line selection</remarks>
	public class TableControl : ScrollPane
	{
		public TableControl()
		{
			// it's so slow ....
			DoubleBuffered = false;
			
			oldsize = Size;
			BackColor = SystemColors.Window;
			body = new TableBody(this);
			colHeader = new TableColumnHeader(this);
			rowHeader = new TableRowHeader(this);
		}
		Size oldsize;

		public bool AllowRowsResizing = false, AllowColumnsResizing = true;
		TableRowHeader rowHeader;
		public TableRowHeader RowHeader
		{
			get { return rowHeader; }
			set
			{
				if(rowHeader != null)
					rowHeader.Table = null;
				rowHeader = value;
				if(rowHeader != null)
					rowHeader.Table = this;
			}
		}
		TableColumnHeader colHeader;
		public TableColumnHeader ColHeader
		{
			get { return colHeader; }
			set
			{
				if(colHeader != null)
					colHeader.Table = null;
				colHeader = value;
				if(colHeader != null)
					colHeader.Table = this;
			}
		}
		ICornerRenderer corner;
		public ICornerRenderer Corner
		{
			get { return corner; }
			set { corner = value; }
		}

		bool showRow = true, showCol = true;
		public bool ShowRowHeader
		{
			get 
			{
				if(RowHeader == null)
					return false;
				return showRow;
			}
			set 
			{
				showRow = value;
			}
		}
		public bool ShowColHeader
		{
			get 
			{
				if(ColHeader == null)
					return false;
				return showCol;
			}
			set 
			{
				showCol = value;
			}
		}

		TableBody body;
		public TableBody Body
		{
			get { return body; }
			set 
			{
				if(value == null)
					throw new NullReferenceException();
				body.Table = null;
				body = value;
				body.Table = this;
			}
		}

		ITableModel model;
		public virtual ITableModel Model
		{
			get { return model; }
			set
			{
				Editor.CancelEditing();
				if(model != null)
					model.Changed -= new TableEventHandler(HModelChanged);
				model = value;
				UpdateSizeStuff();
				if(model != null)
					model.Changed += new TableEventHandler(HModelChanged);
				SelectedRow = -1;
				Invalidate();
			}
		}
		protected virtual void HModelChanged(TableModelEvent tme)
		{
			int row0 = tme.FirstRow, row1 = tme.LastRow;
			switch(tme.EventType)
			{
				default:
				case TableModelEvent.EType.All:
					Editor.CancelEditing();
					UpdateSizeStuff();
					Invalidate();
					break;
				case TableModelEvent.EType.Structure:
					// Reset()is ony call here and not in
					// Model = value,
					// so the user could programatically set cell's size
					// and expect it to work
					Editor.CancelEditing();
					Body.Reset(); 
					UpdateSizeStuff();
					SelectedRow = -1;
					Invalidate();
					break;
				case TableModelEvent.EType.RowInserted:
					Editor.StopEditing(); // simple, no hassle
					Body.InsertRows(row0, 1+row1-row0);
					Rectangle r0 = LineBounds(row0);
					//Rectangle r1 = LineBounds(row1);
					Rectangle r2 = LineBounds(model.NumRows-1);
					Rectangle r = new Rectangle(r0.X, r0.Y, r0.Width, r2.Bottom-r0.Top);
					// can't do that, because of alternate line color
					//ScrollArea(r, new Point(0, r1.Bottom - r0.Top));
					Invalidate(r);
					UpdateSizeStuff();
					if(SelectedRow >= row0)
						SelectedRow += 1+row1-row0;
					break;
				case TableModelEvent.EType.RowRemoved:
					Editor.StopEditing(); // simple, no hassle
					r0 = LineBounds(row0);
					//r1 = LineBounds(row1);
					r2 = GetCellBounds(Model.NumRows+row1-row0, 0); // line bounds is only for real row
					Body.DeleteRows(row0, 1+row1-row0);
					r = new Rectangle(r0.X, r0.Y, r0.Width, r2.Bottom-r0.Top);
					// can't do that, because of alternate line color
					//ScrollArea(r, new Point(0, r0.Top - r1.Bottom));
					Invalidate(r);
					UpdateSizeStuff();
					if(SelectedRow > row1)
						SelectedRow -= 1+row1-row0;
					else if(SelectedRow >=row0 && SelectedRow <= row1)
						SelectedRow = -1;
					break;
				case TableModelEvent.EType.RowUpdated:
					RedrawLines(tme.FirstRow, tme.LastRow, tme.Column);
					break;
			}
		}
		TableEditorManager editorManager;
		public TableEditorManager Editor
		{
			get 
			{
				if(editorManager == null)
				{
					editorManager = CreateEditorManager();
					editorManager.table = this;
				}
				return editorManager;
			}
			set
			{
				if(editorManager != null) 
				{
					editorManager.StopEditing();
					editorManager.table = null;
				}
				editorManager = value;
				if(editorManager != null)
					editorManager.table = this;
			}
		}
		protected virtual TableEditorManager CreateEditorManager()
		{
			return new TableEditorManager();
		}

		TableRendererManager rendererManager;
		public TableRendererManager Renderers
		{
			get 
			{
				if(rendererManager == null)
				{
					rendererManager = CreateRendererManager();
					rendererManager.table = this;
				}
				return rendererManager;
			}
			set
			{
				if(rendererManager != null) 
					rendererManager.table = null;
				rendererManager = value;
				if(rendererManager != null)
					rendererManager.table = this;
			}
		}
		protected virtual TableRendererManager CreateRendererManager()
		{
			return new TableRendererManager();
		}

		protected override void OnResize(EventArgs e)
		{
			Resized(oldsize, Size);
			oldsize = Size;
			base.OnResize(e);
		}
		// redraw the border lines after resize
		void Resized(Size olds, Size news)
		{
			if(olds.Width > news.Width)
				Invalidate(new Rectangle(news.Width-1, 0, 1, news.Height));
			else if(olds.Width < news.Width)
				Invalidate(new Rectangle(olds.Width-1, 0, 1, news.Height));

			if(olds.Height > news.Height)
				Invalidate(new Rectangle(0, news.Height-1, news.Width, 1));
			else if(olds.Height < news.Height)
				Invalidate(new Rectangle(0, olds.Height-1, news.Width, 1));
		}

		/// <summary>
		/// recalculate the scrollable size
		/// </summary>
		public void UpdateSizeStuff()
		{
			if(Editor != null)
				Editor.StopEditing();
			Point p = BodyOffset;
			Size aSize = Body.Size;
			aSize.Width += p.X;
			aSize.Height += p.Y;
			ScrollableSize = aSize;

			aSize = Body.GetCellSize(-1, -1);
			HScrollBar.SmallChange = aSize.Width;
			VScrollBar.SmallChange = aSize.Height;
		}
		/// <summary>
		/// tell the control to recalculate its size and repaint itself
		/// the number of row or columns might have changed, and so did 
		/// some values ..
		/// </summary>
		public void ModelUpdated()
		{
			if(Editor != null)
				Editor.StopEditing();
			UpdateSizeStuff();
			Invalidate();
		}

		int selected = -1;
		/// <summary>
		/// <summary>
		/// the index of the selected row or -1
		/// </summary>
		public virtual int SelectedRow
		{
			get { return selected; }
			set
			{
				if(!CanSelectLine)
					value = -1;
				if(Model == null)
					return;
				if(value < 0 || value >= Model.NumRows)
					value = -1;
				if(value == selected)
					return;
				int old = selected;
				selected = value;

				if(old > -1) 
					RedrawLine(old/*, true*/);
				ScrollToVisible(selected);
				RedrawLine(selected/*, true*/);

				// fire event
				if(RowSelected != null)
					RowSelected(selected);
			}
		}

		/// <summary>
		/// simple select flag. if it happend selection becomes more 
		/// complicated should make a SelectionModel, like for the JTable
		/// </summary>
		public virtual bool CanSelectLine
		{
			get { return canSelect; }
			set
			{
				canSelect = value;
				if(! canSelect)
					SelectedRow = -1;
			}
		}
		bool canSelect = true;

		/// <summary>
		/// don't call parent method and do personal treatment
		/// </summary>
		/// <param name="e"></param>
		protected override void OnKeyDown(KeyEventArgs e)
		{
			ApplicationEx.ClearCurrent();
			int sel = SelectedRow;

			switch(e.KeyCode)
			{
				case Keys.Up:
					if(sel > -1)
					{
						if(SelectedRow > 0)
							SelectedRow --;
					}
					else 
						base.OnKeyDown(e);
					break;
				case Keys.Down:
					if(sel > -1) 
					{
						if(Model != null && sel < Model.NumRows-1)
							SelectedRow ++;
					}
					else 
						base.OnKeyDown(e);
					break;
				default:
					base.OnKeyDown(e);
					break;
			}
		}

		/// <summary>
		/// redraw some lines when data have changed...
		/// should be done automatically for model firing proper events
		/// </summary>
		/// <param name="row0">the first modified row</param>
		/// <param name="row1">the last modified row</param>
		public void RedrawLines(int row0, int row1)
		{
			RedrawLines(row0, row1, -1);
		}
		/// <summary>
		/// redraw some lines when data have changed...
		/// should be done automatically for model firing proper events.
		/// </summary>
		/// <param name="row0">the first modified row</param>
		/// <param name="row1">the last modified row</param>
		/// <param name="col">the column affected or -1 for all the row</param>
		public void RedrawLines(int row0, int row1, int col)
		{
			Rectangle r0 = LineBounds(row0);
			Rectangle r1 = LineBounds(row1);
			Rectangle r = new Rectangle(r0.X, r0.Y, r0.Width, r1.Bottom - r0.Top);
			if(col > -1)
			{
				Rectangle rc = GetCellBounds(row0, col);
				r.X = rc.X;
				r.Width = rc.Width;
			}
			if(r.IntersectsWith(new Rectangle(0, 0, Width, Height)))
				Invalidate(r);
		}
		/// <summary>
		/// mark a row as dirty
		/// </summary>
		/// <param name="row"></param>
		public void RedrawLine(int row)
		{
			if(Model == null || row < 0 || row >= Model.NumRows)
				return;
			Rectangle r = LineBounds(row);
			r.Intersect(ClientRectangle);
			if(r.Width > 0 && r.Height > 0)
				Invalidate(r);
		}
		Rectangle LineBounds(int row)
		{
			//if(Model == null || row < 0 || row >= Model.NumRows)
			//	return new Rectangle(0,0,0,0);
			Rectangle r = GetCellBounds(row, 0);
			r.X = 0;
			r.Width = ScrollableSize.Width+Width;
			return r;
		}

		bool showRowGrid = false, showColGrid = true;
		public bool ShowRowLine
		{
			get { return showRowGrid; }
			set 
			{
				showRowGrid = value;
			}
		}
		public bool ShowColLine
		{
			get { return showColGrid; }
			set 
			{
				showColGrid = value;
			}
		}

		/// <summary>
		/// return the offset at which the body is drawn
		/// </summary>
		public Point BodyOffset
		{
			get
			{
				int x = 0, y = 0;
				if(ShowRowHeader)
					x = RowHeader.Width;
				if(ShowColHeader)
					y = ColHeader.Height;
				return new Point(x, y);
			}
		}
		/// <summary>
		/// find the cell at the given location in control coordinate
		/// </summary>
		/// <param name="p">the location in the control</param>
		/// <param name="row">the row of the cell</param>
		/// <param name="col">the column of the cell</param>
		public void CellAtPoint(Point p, out int row, out int col)
		{
			GetVirtualCellAt(p, out row, out col);

			// bounds checking
			// -2 != -1 value for headers
			if(row >= Model.NumRows)
				row = -2;
			if(col >= Model.NumColumns)
				col = -2;
		}
		/// <summary>
		/// because the size of cells model is infinite, this return
		/// the row and column for any point, even if there is no cell there
		/// </summary>
		/// <param name="p">the location in the control</param>
		/// <param name="row">the row of the cell</param>
		/// <param name="col">the column of the cell</param>
		public void GetVirtualCellAt(Point p, out int row, out int col)
		{
			if(Model == null)
			{
				row = -1;
				col = -1;
				return;
			}
			bool rowHeader, colHeader;
			if(ShowColHeader) 
			{
				colHeader = p.Y <= ColHeader.Height;
				p.Y -= ColHeader.Height;
			}
			else
				colHeader = false;
			if(ShowRowHeader)
			{
				rowHeader = p.X <= RowHeader.Width;
				p.X -= RowHeader.Width;
			}
			else 
				rowHeader = false;
			Point dp = ScrollableLocation;
			p.X -= dp.X;
			p.Y -= dp.Y;
			Body.CellAtPoint(p, out row, out col);
			if(colHeader)
				row = -1;
			if(rowHeader)
				col = -1;
		}
		public Rectangle GetCellBounds(int row, int col)
		{
			Point p = Body.GetCellOffset(row, col);

			Point dp = ScrollableLocation;
			p.X += dp.X;
			p.Y += dp.Y;
			if(ShowColHeader)
				p.Y += ColHeader.Height;
			if(ShowRowHeader)
				p.X += RowHeader.Width;

			Size s = Body.GetCellSize(row, col);

			return new Rectangle(p.X, p.Y, s.Width, s.Height);
		}

		/// <summary>
		/// the pixel distance to the headers knob at which the mouse event
		/// is a columns/rows resize event.
		/// </summary>
		const int MouseSensibility = 4;

		/// <summary>
		/// a flag for mouse drag event, which says wether or not 
		/// </summary>
		class HeaderDragInfo
		{
			public HeaderDragInfo()
			{
				previousSize = -1;
			}
			public bool dragColumn;
			public int draggedHeader;
			public int previousSize;
		}
		HeaderDragInfo dragInfo;

		protected override void OnMouseMove(MouseEventArgs e)
		{
			base.OnMouseMove(e);
			if(dragInfo != null)
			{
				int row = -1, col = -1;
				if(dragInfo.dragColumn)
					col = dragInfo.draggedHeader;
				else 
					row = dragInfo.draggedHeader;
				Rectangle r = GetCellBounds(row, col);

				int newWidth;
				if(dragInfo.dragColumn)
					newWidth = e.X - r.Left;
				else
					newWidth = e.Y - r.Top;
				ResizeHeaderTo(newWidth);
			}
		}
		protected override void OnMouseUp(MouseEventArgs e)
		{
			base.OnMouseUp (e);
			if(dragInfo != null)
			{
				if(dragInfo.dragColumn)
					Body.SetColWidth(dragInfo.draggedHeader, dragInfo.previousSize);
				else
					Body.SetRowHeight(dragInfo.draggedHeader, dragInfo.previousSize);
				UpdateSizeStuff();
				Refresh();

				dragInfo = null;
				return;
			}

			int row, col;
			GetVirtualCellAt(new Point(e.X, e.Y), out row, out col);
			Rectangle r = GetCellBounds(row, col);
			TableClickEvent mouseUp = new TableClickEvent(e, this, r, row, col);
			if(mouseDown!=null 
				&& !mouseDown.Consumed
				&& mouseUp.Row == mouseDown.Row
				&& mouseUp.Col == mouseDown.Col)
			{
				if(TableClickUp != null)
					TableClickUp(this, mouseUp);

				if(!mouseUp.Consumed
					&& Model != null 
					&& mouseUp.Row > -1
					&& mouseUp.Col > -1
					&& Model.IsEditable(mouseUp.Row, mouseUp.Col))
					Editor.StartEditing(mouseUp.Row, mouseUp.Col);
			}
		}
		protected virtual void ResizeHeaderTo(int newSize)
		{
			if(newSize < 2 * MouseSensibility)
				newSize = 2 * MouseSensibility;
			if(newSize == dragInfo.previousSize)
				return;

			int row = -1, col = -1;
			if(dragInfo.dragColumn)
				col = dragInfo.draggedHeader;
			else
				row = dragInfo.draggedHeader;
			Rectangle r = GetCellBounds(row, col);

			// redraw old-line (XOR erasing)
			Rectangle r2;
			if(dragInfo.previousSize > -1)
			{
				if(dragInfo.dragColumn)
					r2 = new Rectangle(r.Left+dragInfo.previousSize, 0, 1, Height);
				else
					r2 = new Rectangle(0, r.Top+dragInfo.previousSize, Width, 1);
				XORPaint(r2, Color.Red);
			}

			// draw newLine
			if(dragInfo.dragColumn)
				r2 = new Rectangle(r.Left+newSize, 0, 1, Height);
			else
				r2 = new Rectangle(0, r.Top+newSize, Width, 1);
			XORPaint(r2, Color.DarkRed);

			dragInfo.previousSize = newSize;

			if(dragInfo.dragColumn && ColumnResized != null)
				ColumnResized(col, newSize);
			else if(! dragInfo.dragColumn && RowResized != null)
				RowResized(row, newSize);
		}
		public delegate void ColumnResizedHandler(int col, int newSize);
		public delegate void RowResizedHandler(int row, int newSize);
		public event ColumnResizedHandler ColumnResized;
		public event RowResizedHandler RowResized;

		public event TableClickEventHandler TableClick;
		/// <summary>
		/// this event is triggered if the mouse up and down events
		/// are in the same existing (row &amp; col > -2) cell
		/// </summary>
		public event TableClickEventHandler TableClickUp;

		TableClickEvent mouseDown;
		protected override void OnMouseDown(MouseEventArgs e)
		{
			if(Editor != null)
				Editor.StopEditing();
			mouseDown = null;

			int row, col;
			GetVirtualCellAt(new Point(e.X, e.Y), out row, out col);
			int nrow = 0, ncol = 0;
			if(Model != null)
			{
				nrow = Model.NumRows;
				ncol = Model.NumColumns;
			}
			Rectangle r = GetCellBounds(row, col);

			Point p = BodyOffset;
			Point mouse = new Point(e.X, e.Y);
			bool consumed = false;

			if(e.X < p.X && e.Y < p.Y)
			{
			}
			else if(e.Y < p.Y && e.X > p.X)
			{
				if(AllowColumnsResizing && r.Right - e.X < MouseSensibility && col > -1)
				{
					dragInfo = new HeaderDragInfo();
					dragInfo.dragColumn = true;
					dragInfo.draggedHeader = col;
					ResizeHeaderTo(e.X - r.Left);
					consumed = true;
				}
				else if(AllowColumnsResizing && e.X - r.Left < MouseSensibility && col > 0)
				{
					dragInfo = new HeaderDragInfo();
					dragInfo.dragColumn = true;
					dragInfo.draggedHeader = col-1;
					ResizeHeaderTo(e.X - r.Left + Body.GetCellSize(row, col-1).Width);
					consumed = true;
				}
			}
			else if(e.X < p.X && e.Y > p.Y)
			{
				if(AllowRowsResizing && r.Bottom - e.Y < MouseSensibility && row > -1)
				{
					dragInfo = new HeaderDragInfo();
					dragInfo.dragColumn = false;
					dragInfo.draggedHeader = row;
					ResizeHeaderTo(e.Y - r.Top);
				}
				else if(AllowRowsResizing && e.Y - r.Top < MouseSensibility && row > 0)
				{
					dragInfo = new HeaderDragInfo();
					dragInfo.dragColumn = false;
					dragInfo.draggedHeader = row-1;
					ResizeHeaderTo(e.Y - r.Top + Body.GetCellSize(row-1, col).Height);
				}
				else
				{
					SelectedRow = row>-1 && row<nrow ? row : -1;
				}
			}
			else if(row > -1 && row < nrow && col < ncol)
			{
				if(row != SelectedRow)
				{
					SelectedRow = row;
					if(Editor.Editing)
						Editor.StopEditing();
				}
			}
			else
			{
				SelectedRow = -1;
			}

			base.OnMouseDown (e);
			mouseDown = new TableClickEvent(e, this, r, row, col);
			if(TableClick != null && !consumed)
				TableClick(this, mouseDown);
		}
		public virtual void ScrollToVisible(int row)
		{
			if(Model == null || row < 0 || row >= Model.NumRows)
				return;

			Point p = Body.GetCellOffset(row, 0);
			Size s = Body.GetCellSize(row, 0);

			Rectangle r = new Rectangle(0, p.Y, ScrollableSize.Width, s.Height);
			ScrollToVisible(r);
		}
		public override void ScrollToVisible(Rectangle r)
		{
			Size sSize = ScrollableSize;
			r.Intersect(new Rectangle(0,0,sSize.Width,sSize.Height));

			Size view = InsideRect.Size;
			Point pb = BodyOffset;
			view.Width -= pb.X;
			view.Height -= pb.Y;

			Point p = ScrollableLocation;
			Point p2 = p;

			ScrollToRectVisible(view, r, ref p);

			if(p2.X != p.X || p2.Y != p.Y)
				ScrollableLocation = p;
		}

		protected override void DoubleBufferedPaintBackground(PaintEventArgs e)
		{
			// do nothing ..
		}
		protected override void DoubleBufferedPaint(PaintEventArgs e)
		{
			Rectangle clip = e.ClipRectangle;
			int x = 0, y = 0;
			if(ShowColHeader) 
				y = ColHeader.Size.Height - 1; // -1, because of border
			if(ShowRowHeader) 
				x = RowHeader.Size.Width - 1; // -1, because of border

			// body....
			{ 
				Rectangle r = new Rectangle(x, y, Width, Height);
				r.Intersect(clip);
				e.Graphics.Clip = new Region(r);
				if(r.Width > 0 && r.Height > 0)
					Body.Paint(new PaintEventArgs(e.Graphics, r));
			}

			// headers ...
			if(ShowColHeader) 
			{
				Rectangle r = new Rectangle(0, 0, Width, y+1);
				r.Intersect(clip);
				e.Graphics.Clip = new Region(r);
				if(r.Width > 0 && r.Height > 0)
				{
					e.Graphics.FillRectangle(AppDefaults.WhiteBrush, r);
					ColHeader.Paint(new PaintEventArgs(e.Graphics, r));
				}
			}
			if(ShowRowHeader) 
			{
				Rectangle r = new Rectangle(0, 0, x+1, Height);
				r.Intersect(clip);
				e.Graphics.Clip = new Region(r);
				if(r.Width > 0 && r.Height > 0)
				{
					e.Graphics.FillRectangle(AppDefaults.WhiteBrush, r);
					RowHeader.Paint(new PaintEventArgs(e.Graphics, r));
				}
			}
			if(ShowColHeader && ShowRowHeader)
			{
				Rectangle r = new Rectangle(0, 0, x+1, y+1);
				Rectangle rc = r;
				r.Intersect(clip);
				e.Graphics.Clip = new Region(r);
				if(r.Width > 0 && r.Height > 0)
				{
					e.Graphics.FillRectangle(AppDefaults.WhiteBrush, r);
					if(Corner != null)
						Corner.PaintCorner(new PaintEventArgs(e.Graphics, r), this, rc);
				}
			}

			// header & border lines ...
			{
				Rectangle r = clip;
				e.Graphics.Clip = new Region(r);
				Size bSize = Body.Size;
				r.Intersect(new Rectangle(x, y, Width, Height));
				Pen pen = AppDefaults.BorderPen;
				if(ShowColHeader) 
					e.Graphics.DrawLine(pen, r.Left, y+1, r.Right, y+1);
				if(ShowRowHeader) 
					e.Graphics.DrawLine(pen, x+1, r.Top, x+1, r.Bottom);
				e.Graphics.DrawRectangle(pen, 0, 0, Width-1, Height-1);
			}
		}
		public virtual void GetBordingCells(Rectangle r, out int rowMin, out int rowMax, out int colMin, out int colMax)
		{
			Point p = ScrollableLocation;
			r.X -= p.X;
			r.Y -= p.Y;
			if(ShowColHeader)
				r.Y -= ColHeader.Height;
			if(ShowRowHeader)
				r.X -= RowHeader.Width;
			Body.CellAtPoint(r.Location, out rowMin, out colMin);
			Body.CellAtPoint(new Point(r.Right-1, r.Bottom-1), out rowMax, out colMax);
		}
		public void GetBordingRows(Rectangle r, out int rowMin, out int rowMax)
		{
			int colMin, colMax;
			GetBordingCells(r, out rowMin, out rowMax, out colMin, out colMax);
		}
		public void GetBordingColumns(Rectangle r, out int colMin, out int colMax)
		{
			int rowMin, rowMax;
			GetBordingCells(r, out rowMin, out rowMax, out colMin, out colMax);
		}
		public override Point ScrollableLocation
		{
			set
			{
				if(Editor.Editing)
					Editor.StopEditing();
				base.ScrollableLocation = value;
			}
		}

		protected override void MovedBy(int dx, int dy)
		{
			if(Editor.Editing) 
				Editor.StopEditing();

			if(dx == 0 && dy == 0)
				return;
			//ApplicationEx.DoEvents(WinAPI.WM_PAINT);
			//ApplicationEx.ClearCurrent();

			int w = Width, h = Height;
			int x = 0, y = 0;
			Rectangle inside = new Rectangle(1, 1, w-2, h-2);
			inside.Intersect(InsideRect);

			ITableModel m = Model;
			int nr = 0;
			int nc = 0;
			if(m != null)
			{
				nr = m.NumRows;
				nc = m.NumColumns;
			}
			Rectangle cellsRect = GetCellBounds(nr, nc);
			cellsRect = new Rectangle(0, 0, cellsRect.X, cellsRect.Y);
			cellsRect.Width += dx<0 ? -dx : dx;
			cellsRect.Height += dy<0 ? -dy : dy;
			inside.Intersect(cellsRect);

			if(ShowColHeader) 
				y = ColHeader.Size.Height;
			if(ShowRowHeader) 
				x = RowHeader.Size.Width;

			if(dx == 0)
			{
				Rectangle r = new Rectangle(1, y+1, w-1, h-1);
				r.Intersect(inside);
				ScrollArea(r, new Point(dx, dy));
			}
			else if(dy == 0)
			{
				Rectangle r = new Rectangle(x+1, 1, w-1, h-1);
				r.Intersect(inside);
				ScrollArea(r, new Point(dx, dy));
			}
			else
			{
				Rectangle r = new Rectangle(x+1, y+1, w-1, h-1);
				r.Intersect(inside);
				ScrollArea(r, new Point(dx, dy));

				if(ShowColHeader)
				{
					r = new Rectangle(x, 0, w, y);
					r.Intersect(inside);
					ScrollArea(r, new Point(dx, 0));
				}

				if(ShowRowHeader)
				{
					r = new Rectangle(0, y, x, h);
					r.Intersect(inside);
					ScrollArea(r, new Point(0, dy));
				}
			}
		}

		#region -- events --
		public delegate void RowSelectedHandler(int row);
		/// <summary>
		/// this is fired whenever the selected row was changed.
		/// when nothing is selected the argument is -1.
		/// </summary>
		public event RowSelectedHandler RowSelected;
		#endregion

		private void TableControl_ParentChanged(object sender, EventArgs e)
		{
			MessageBox.Show("Parent Changed");
		}
	}
}
