#region -- BSD Licence --
/*
Copyright (c) 2003, 5D Services
http://www.ihook.cc
OpenSource@ihook.cc
lloyd@5dservices.com.au
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. 

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution. 

Neither the name of "5D Services" nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
#endregion
using System;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Windows.Forms;

using ihook.PocketWidgets.Editors;
//using ihook.PocketWidgets.Table.Editors;
using ihook.PocketWidgets.Utils;

namespace ihook.PocketWidgets.Table
{
	/// <summary>
	/// Handle the editing of a table
	/// </summary>
	public class TableEditorManager
	{
		ITableEditor editor;
		int row = -1, col = -1;
		bool valueChanged;

		public TableEditorManager()
		{
			SetupDefaultEditors();
		}
		protected virtual void SetupDefaultEditors()
		{
			this[typeof(DateTime)] = new TableEditor(new DateEditor());
			this[typeof(bool)] = new BoolTableEditor();
			this[typeof(decimal)] = new TableEditor(new NumEditor(NumType.Decimal));
			this[typeof(double)] = new TableEditor(new NumEditor(NumType.Double));
			this[typeof(float)] = new TableEditor(new NumEditor(NumType.Single));
			this[typeof(short)] = new TableEditor(new UpDownNumEditor(NumType.Short));
			this[typeof(int)] = new TableEditor(new UpDownNumEditor(NumType.Int));
			this[typeof(long)] = new TableEditor(new UpDownNumEditor(NumType.Long));
		}

		protected internal TableControl table;
		public TableControl Table
		{
			get { return table; }
		}

		/// <summary>
		/// stop the editing and apply changes (if any)
		/// </summary>
		public virtual void StopEditing()
		{
			StopEditing(true);
		}
		/// <summary>
		/// this stop the editing and dismiss the changes
		/// </summary>
		public virtual void CancelEditing()
		{
			StopEditing(false);
		}
		void StopEditing(bool update)
		{
			if(editor == null)
				return;

			editor.ValueChanged -= new EventHandler(HValueChanged);

			if(table != null)
				table.Controls.Remove(editor.Control);

			if(update && valueChanged)
				ApplyChange();

			editor = null;
			row = col = -1;
		}
		void ApplyChange()
		{
			if(editor != null
				&& table != null 
				&& table.Model != null 
				&& row > -1 
				&& col > -1)
					table.Model[row, col] = editor.Value;
		}

		public bool Editing { get { return row != -1 && editor != null; } }
		public int EditingRow { get { return row; } }
		public int EditingCol { get { return col; } }

		#region -- find the editor ! --
		/// <summary>
		/// return the editor for a given cell. the default is 
		/// to ask for the column editor (<c>this[col]</c>)
		/// </summary>
		public virtual ITableEditor this[int row, int col]
		{
			get { return this[col]; }
		}
		/// <summary>
		/// return the editor for a colum. check first if there
		/// is an editor set for this column, if there is none
		/// search for an editor for type of this column, or null
		/// if there is not ITableModel
		/// </summary>
		public virtual ITableEditor this[int col]
		{
			get
			{
				if(colEditor != null)
				{
					ITableEditor te = (ITableEditor) colEditor[col];
					if(te != null)
						return te;
				}
				TableControl tc = Table;
				if(tc == null)
					return null;
				ITableModel tm = tc.Model;
				if(tm == null)
					return null;
				return this[tm.GetColumnType(col)];
			}
			set
			{
				if(value == null)
				{
					if(colEditor == null)
						return;
					colEditor.Remove(col);
				}
				else
				{
					if(colEditor == null)
						colEditor = new Hashtable();
					colEditor[col] = value;
				}
			}
		}
		Hashtable colEditor;

		/// <summary>
		/// get or set the editor to use for a given type,
		/// all others default indexed accesssor call this one.
		/// </summary>
		/// <remarks>if there is no specific editor set for
		/// this type, it return the default editor</remarks>
		public virtual ITableEditor this[Type t]
		{
			get
			{
				if(t == null)
					return DefaultEditor;
				ITableEditor te = (ITableEditor) editors[t];
				if(te == null)
					return DefaultEditor;
				return te;
			}
			set
			{
				if(t == null)
					return;
				if(value == null)
					editors.Remove(t);
				else
					editors[t] = value;
			}
		}
		Hashtable editors = new Hashtable();

		/// <summary>
		/// the default editor when none is set. by default its
		/// a TableTextEditor, this property never return null
		/// </summary>
		public virtual ITableEditor DefaultEditor
		{
			get 
			{
				if(defaultEditor == null)
					defaultEditor = new TableEditor(new TextEditor());
				return defaultEditor;
			}
			set { defaultEditor = value; }
		}
		ITableEditor defaultEditor;
		#endregion

		#region editing !
		/// <summary>
		/// start editing a row by creating an editor and put it in the cell.
		/// this method check first if the cell is editable, according to the model
		/// </summary>
		/// <param name="row"></param>
		/// <param name="col"></param>
		/// <returns>return wether or not the editing was started
		/// successfully</returns>
		public virtual void StartEditing(int aRow, int aCol)
		{
			valueChanged = false;
			StopEditing();
			if(aRow < 0 || aCol < 0) 
				return;
			if(table == null)
				return;
			ITableModel tm = table.Model;
			if(tm == null)
				return;
			if(!tm.IsEditable(aRow, aCol))
				return;
			editor = this[aRow, aCol];
			if(editor == null)
				return;

			row = aRow;
			col = aCol;

			bool edit = editor.Prepare(row, col, table);
			Control c = editor.Control;
			if(edit && c != null)
			{
				editor.ValueChanged += new EventHandler(HValueChanged);
				c.Bounds = table.GetCellBounds(row, col);
				table.Controls.Add(c);
				c.Focus();
			}
			else
			{
				if(edit)
					ApplyChange();
				row = -1;
				col = -1;
				editor = null;
			}
		}
		void HValueChanged(object sender, EventArgs ea)
		{
			valueChanged = true;
		}
		#endregion
	}
}
