using System;
using System.Drawing;
using System.Collections;
using System.Windows.Forms;
using System.Data;

using ihook.PocketWidgets;

namespace ihookDB.Widgets.Paint
{
	/// <summary>
	/// Summary description for PaintArea.
	/// </summary>
	public class PaintArea : DoubleBufferedControl
	{
		PaintForm master;
		public PaintArea(PaintForm master)
		{
			this.master = master;
		}

		Image model;
		Bitmap backImg;
		Graphics backGraph;
		public void ReInit()
		{
			if(backGraph != null)
			{
				backGraph.Dispose();
				backGraph = null;
			}
			if(backImg != null)
			{
				backImg.Dispose();
				backImg = null;
			}
			int w = Width - 2;
			int h = Height - 2;
			if(w < 1)
				w = 1;
			if(h < 1)
				h = 1;
			backImg = new Bitmap(w, h);
			backGraph = Graphics.FromImage(backImg);

			Brush brush = new SolidBrush(Color.White);
			backGraph.FillRectangle(brush, 0, 0, w, h);
			brush.Dispose();

			if(model != null)
				backGraph.DrawImage(model, 0, 0);

			Invalidate();
		}
		public Bitmap Image
		{
			get 
			{ 
				if(backImg == null)
				{
					int w = Width, h = Height;
					if(w < 1)
						w = 1;
					if(h < 1)
						h = 1;
					return new Bitmap(w, h);
				}
				return new Bitmap(backImg);
			}
		}
		public Image Model
		{
			get { return model; }
			set
			{
				model = value;
				ReInit();
			}
		}
		public Graphics ImageGraphics 
		{ 
			get 
			{ 
				if(backGraph == null)
					ReInit();
				return backGraph; 
			}
		}

		protected override void OnResize(EventArgs e)
		{
			base.OnResize (e);
			ReInit();
		}

		protected override void DoubleBufferedPaintBackground(PaintEventArgs e)
		{
			// nothing....
		}
		protected override void DoubleBufferedPaint(PaintEventArgs e)
		{
			if(backImg == null)
				ReInit();

			e.Graphics.DrawImage(backImg, 1, 1);

			if(Capture)
				switch(info.tool)
				{
					case PaintForm.Tools.Pen:
					case PaintForm.Tools.SprayCan:
						break;
					case PaintForm.Tools.Line:
						Line(e.Graphics, lastPoint, mouseDown);
						break;
					case PaintForm.Tools.Ellipse:
						Ellipse(e.Graphics, lastPoint, mouseDown);
						break;
					case PaintForm.Tools.Rectangle:
						Rectangle(e.Graphics, lastPoint, mouseDown);
						break;
				}

			e.Graphics.DrawRectangle(AppDefaults.TextPen, 0, 0, Width-1, Height-1);
		}

		bool dragging = false;
		Point mouseDown;
		Point lastPoint;
		class DrawInfo
		{
			public Color color;
			public int width;
			public PaintForm.Tools tool;
			public bool fill;

			public Pen pen;
			public Brush brush;
		}
		DrawInfo info;
		void GetInfo()
		{
			if(info == null)
				info = new DrawInfo();

			info.tool = master.Tool;
			info.fill = master.checkFill.Checked;

			info.width = master.tThick.Value;
			if(info.width < 1)
				info.width = 1;

			Color color = master.cChooser.SelectedColor;
			if(color != info.color)
			{
				info.color = color;
				if(info.pen != null)
					info.pen.Dispose();
				info.pen = new Pen(info.color);
				if(info.brush != null)
					info.brush.Dispose();
				info.brush = new SolidBrush(color);
			}
		}
		protected override void OnMouseDown(MouseEventArgs e)
		{
			base.OnMouseDown (e);
			lastPoint = mouseDown = new Point(e.X-1, e.Y-1);
			Focus();
			Capture = true;
			dragging = true;

			GetInfo();
			switch(info.tool)
			{
				case PaintForm.Tools.Pen:
					if(info.width < 2)
						Point(lastPoint.X, lastPoint.Y);
					else
						backGraph.FillEllipse(info.brush, 
							(int) Math.Ceiling(lastPoint.X-info.width/2.0), 
							(int) Math.Ceiling(lastPoint.Y-info.width/2.0), 
							info.width, info.width);
					Refresh();
					break;
				case PaintForm.Tools.SprayCan:
					Spray(lastPoint);
					Refresh();
					break;
				case PaintForm.Tools.Line:
				case PaintForm.Tools.Ellipse:
				case PaintForm.Tools.Rectangle:
					break;
			}
		}
		protected override void OnMouseMove(MouseEventArgs e)
		{
			base.OnMouseMove (e);
			if(!Capture)
				return;

			Point p = new Point(e.X-1, e.Y-1);
			switch(info.tool)
			{
				case PaintForm.Tools.Pen:
					Line(backGraph, p, lastPoint);
					break;
				case PaintForm.Tools.SprayCan:
					Spray(p);
					break;
				case PaintForm.Tools.Line:
				case PaintForm.Tools.Ellipse:
				case PaintForm.Tools.Rectangle:
					break;
			}
			lastPoint = p;
			Refresh();
		}
		protected override void OnMouseUp(MouseEventArgs e)
		{
			base.OnMouseUp (e);
			if(!dragging) // for some reason Capture is already reset
				return;

			Capture = false;
			dragging = false;
			lastPoint = new Point(e.X, e.Y);
			switch(info.tool)
			{
				case PaintForm.Tools.Pen:
					break;
				case PaintForm.Tools.SprayCan:
					break;
				case PaintForm.Tools.Line:
					Line(backGraph, mouseDown, lastPoint);
					Invalidate();
					break;
				case PaintForm.Tools.Ellipse:
					Ellipse(backGraph, mouseDown, lastPoint);
					Invalidate();
					break;
				case PaintForm.Tools.Rectangle:
					Rectangle(backGraph, mouseDown, lastPoint);
					Invalidate();
					break;
			}
		}

		void Point(int x, int y)
		{
			if(x < backImg.Width 
				&& y < backImg.Height
				&& x > -1
				&& y > -1)
				backImg.SetPixel(x, y, info.color);
		}
		void Line(Graphics g, Point p1, Point p2)
		{
			if(info.width < 2)
			{
				g.DrawLine(info.pen, 
					p1.X, p1.Y,
					p2.X, p2.Y);
				return;
			}

			Point d = new Point(p2.X-p1.X, p2.Y-p1.Y);
			double dx = - d.Y;
			double dy = d.X;
			double m = Math.Sqrt(dx * dx + dy * dy);
			if(m >= 1)
			{
				dx = dx * info.width / m / 2;
				dy = dy * info.width / m / 2;

				Point poly1 = new Point((int)Math.Ceiling(p1.X-dx), (int)Math.Ceiling(p1.Y-dy));
				Point poly2 = new Point((int)Math.Ceiling(p1.X+dx), (int)Math.Ceiling(p1.Y+dy));
				Point poly3 = new Point((int)Math.Ceiling(p2.X+dx), (int)Math.Ceiling(p2.Y+dy));
				Point poly4 = new Point((int)Math.Ceiling(p2.X-dx), (int)Math.Ceiling(p2.Y-dy));

				g.FillPolygon(info.brush, new Point[] {poly1, poly2, poly3, poly4});
				g.FillEllipse(info.brush, GetRectangle(poly1, poly2));
				g.FillEllipse(info.brush, GetRectangle(poly3, poly4));
			}
			else
				g.FillEllipse(info.brush, 
					(int) Math.Ceiling(p1.X-info.width/2.0), 
					(int) Math.Ceiling(p1.Y-info.width/2.0), 
					info.width, info.width);
		}
		Random rand = new Random();
		void Spray(Point p)
		{
			for(int i=0; i<15; i++)
			{
				int x = p.X-info.width/2-1 + rand.Next(info.width);
				int y = p.Y-info.width/2-1 + rand.Next(info.width);
				Point(x, y);
			}
		}
		Rectangle GetRectangle(Point p1, Point p2)
		{
			Rectangle r= new Rectangle();
			if(p1.X < p2.X)
			{
				r.X = p1.X;
				r.Width = p2.X - p1.X;
			}
			else
			{
				r.X = p2.X;
				r.Width = p1.X - p2.X;
			}
			if(p1.Y < p2.Y)
			{
				r.Y = p1.Y;
				r.Height = p2.Y - p1.Y;
			}
			else
			{
				r.Y = p2.Y;
				r.Height = p1.Y - p2.Y;
			}
			return r;
		}
		void Ellipse(Graphics g, Point p1, Point p2)
		{
			Rectangle r = GetRectangle(p1, p2);
			if(info.fill)
			{
				g.FillEllipse(info.brush, r);
			}
			else
			{
				r.Inflate(info.width/2, info.width/2);
				for(int i=0; i<info.width; i++)
				{
					if(r.Width < 1 || r.Height < 1)
						return;
					g.DrawEllipse(info.pen, r);
					r.Inflate(-1, 0);
					if(r.Width < 1 || r.Height < 1)
						return;
					g.DrawEllipse(info.pen, r);
					r.Inflate(0, -1);
				}
			}
		}
		void Rectangle(Graphics g, Point p1, Point p2)
		{
			Rectangle r = GetRectangle(p1, p2);
			if(info.fill)
			{
				g.FillRectangle(info.brush, r);
			}
			else
			{
				r.Inflate(info.width/2, info.width/2);
				for(int i=0; i<info.width; i++)
				{
					if(r.Width < 1 || r.Height < 1)
						return;
					g.DrawRectangle(info.pen, r);
					r.Inflate(-1, -1);
				}
			}
		}
	}
}
